/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [http://neo4j.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.neo4j.cypher

import org.neo4j.cypher.internal.javacompat.NotificationTestSupport.TestFunctions
import org.neo4j.cypher.internal.javacompat.NotificationTestSupport.TestProcedures
import org.neo4j.cypher.internal.util.test_helpers.CypherFunSuite
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_BTREE_INDEX_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_CASE_EXPRESSION
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_COERCION_OF_LIST_TO_BOOLEAN
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_CREATE_CONSTRAINT_ON_ASSERT_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_CREATE_INDEX_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_CREATE_PROPERTY_EXISTENCE_CONSTRAINT_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_DROP_CONSTRAINT_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_DROP_INDEX_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_FUNCTION
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_HEX_LITERAL_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_OCTAL_LITERAL_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_PARAMETER_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_PERIODIC_COMMIT
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_POINTS_COMPARE
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_PROCEDURE
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_PROCEDURE_RETURN_FIELD
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_PROPERTY_EXISTENCE_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_RELATIONSHIP_TYPE_SEPARATOR
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_SELF_REFERENCE_TO_VARIABLE_IN_CREATE_PATTERN
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_SHOW_EXISTENCE_CONSTRAINT_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_SHOW_SCHEMA_SYNTAX
import org.neo4j.graphdb.impl.notification.NotificationCode.DEPRECATED_USE_OF_PATTERN_EXPRESSION
import org.neo4j.graphdb.impl.notification.NotificationCode.LENGTH_ON_NON_PATH
import org.neo4j.graphdb.impl.notification.NotificationDetail
import org.neo4j.graphdb.schema.IndexSettingImpl.FULLTEXT_EVENTUALLY_CONSISTENT
import org.neo4j.graphdb.schema.IndexSettingImpl.SPATIAL_CARTESIAN_MAX
import org.neo4j.graphdb.schema.IndexSettingImpl.SPATIAL_WGS84_MIN
import org.neo4j.kernel.impl.index.schema.GenericNativeIndexProvider
import org.neo4j.kernel.impl.index.schema.fusion.NativeLuceneFusionIndexProviderFactory30
import org.scalatest.BeforeAndAfterAll

abstract class DeprecationAcceptanceTestBase extends CypherFunSuite with BeforeAndAfterAll with DeprecationTestSupport {

  override def beforeAll(): Unit = {
    // Used for testing deprecated procedures
    dbms.registerProcedure(classOf[TestProcedures])
    dbms.registerFunction(classOf[TestFunctions])
    dbms.registerAggregationFunction(classOf[TestFunctions])
  }

  override def afterAll(): Unit = {
    dbms.shutdown()
  }

  // DEPRECATED PROCEDURE THINGS

  test("deprecated procedure calls") {
    val queries = Seq("CALL oldProc()", "CALL oldProc() RETURN 1")
    val detail = NotificationDetail.Factory.deprecatedName("oldProc", "newProc")
    assertNotificationInSupportedVersions(queries, DEPRECATED_PROCEDURE, detail)
  }

  test("non-deprecated procedure calls") {
    val queries = Seq("CALL newProc()", "CALL newProc() RETURN 1")

    assertNoDeprecations(queries)
  }

  test("deprecated procedure result field") {
    val query = "CALL changedProc() YIELD oldField RETURN oldField"
    val detail = NotificationDetail.Factory.deprecatedField("changedProc", "oldField")
    assertNotificationInSupportedVersions(query, DEPRECATED_PROCEDURE_RETURN_FIELD, detail)
  }
  test("non-deprecated procedure result field") {
    val queries = Seq("CALL changedProc() YIELD newField RETURN newField")

    assertNoDeprecations(queries)
  }

  test("deprecated function calls") {
    val queries = Seq(
      "RETURN org.example.com.oldFunc()",
      "MATCH (n) WHERE org.example.com.oldFunc() = 1 RETURN n"
    )
    val detail = NotificationDetail.Factory.deprecatedName("org.example.com.oldFunc", "org.example.com.newFunc")
    assertNotificationInSupportedVersions(queries, DEPRECATED_FUNCTION, detail)
  }

  test("deprecated aggregation function calls") {
    val queries = Seq(
      "UNWIND [1, 2, 3] AS nums RETURN org.example.com.oldAggFunc(nums)",
      "UNWIND [1, 2, 3] AS nums WITH org.example.com.oldAggFunc(nums) AS aggTest RETURN aggTest"
    )
    val detail = NotificationDetail.Factory.deprecatedName("org.example.com.oldAggFunc", "org.example.com.newAggFunc")
    assertNotificationInSupportedVersions(queries, DEPRECATED_FUNCTION, detail)
  }

  test("non-deprecated function calls") {
    val queries = Seq(
      "RETURN org.example.com.newFunc()",
      "MATCH (n) WHERE org.example.com.newFunc() = 1 RETURN n"
    )

    assertNoDeprecations(queries)
  }

  test("non-deprecated aggregation function calls") {
    val queries = Seq(
      "UNWIND [1, 2, 3] AS nums RETURN org.example.com.newAggFunc(nums)",
      "UNWIND [1, 2, 3] AS nums WITH org.example.com.newAggFunc(nums) AS aggTest RETURN aggTest"
    )

    assertNoDeprecations(queries)
  }

  // DEPRECATED INDEX / CONSTRAINT SYNTAX in 4.X

  test("deprecated create index syntax") {
    assertNotificationInSupportedVersions("CREATE INDEX ON :Label(prop)", DEPRECATED_CREATE_INDEX_SYNTAX)
  }

  test("deprecated create btree index syntax") {
    // Note: This index syntax was introduced in 4.X

    // CREATE BTREE INDEX ...
    assertNotificationInSupportedVersions_4_X("CREATE BTREE INDEX FOR (n:Label) ON (n.prop)", DEPRECATED_BTREE_INDEX_SYNTAX)
    assertNotificationInSupportedVersions_4_X("CREATE BTREE INDEX name FOR ()-[r:TYPE]-() ON (r.prop)", DEPRECATED_BTREE_INDEX_SYNTAX)

    // CREATE INDEX ... OPTIONS { <btree options> }
    assertNotificationInSupportedVersions_4_X(
      s"""CREATE INDEX FOR (n:Label) ON (n.prop)
         |OPTIONS {IndexProvider: '${GenericNativeIndexProvider.DESCRIPTOR.name()}'}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
    assertNotificationInSupportedVersions_4_X(
      s"""CREATE INDEX FOR ()-[r:TYPE]-() ON (r.prop)
         |OPTIONS {indexprovider: '${NativeLuceneFusionIndexProviderFactory30.DESCRIPTOR.name()}'}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
    assertNotificationInSupportedVersions_4_X(
      s"""CREATE INDEX FOR (n:Label) ON (n.prop)
         |OPTIONS {IndexConfig: {`${SPATIAL_CARTESIAN_MAX.getSettingName}`: [40, 60]}}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
    assertNotificationInSupportedVersions_4_X(
      s"""CREATE INDEX FOR ()-[r:TYPE]-() ON (r.prop)
         |OPTIONS {indexconfig: {`${SPATIAL_WGS84_MIN.getSettingName}`: [-40, -60]}}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
    assertNoNotificationInSupportedVersions_4_X(
      s"""CREATE INDEX FOR (n:Label) ON (n.prop)
         |OPTIONS {indexconfig: {`${FULLTEXT_EVENTUALLY_CONSISTENT.getSettingName}`: false}}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
  }

  test("deprecated drop index syntax") {
    assertNotificationInSupportedVersions("DROP INDEX ON :Label(prop)", DEPRECATED_DROP_INDEX_SYNTAX)
  }

  test("deprecated drop node key constraint syntax") {
    assertNotificationInSupportedVersions("DROP CONSTRAINT ON (n:Label) ASSERT (n.prop) IS NODE KEY", DEPRECATED_DROP_CONSTRAINT_SYNTAX)
  }

  test("deprecated drop uniqueness constraint syntax") {
    assertNotificationInSupportedVersions("DROP CONSTRAINT ON (n:Label) ASSERT (n.prop) IS UNIQUE", DEPRECATED_DROP_CONSTRAINT_SYNTAX)
  }

  test("deprecated drop node property existence constraint syntax") {
    assertNotificationInSupportedVersions("DROP CONSTRAINT ON (n:Label) ASSERT EXISTS (n.prop)", DEPRECATED_DROP_CONSTRAINT_SYNTAX)
  }

  test("deprecated drop relationship existence constraint syntax") {
    assertNotificationInSupportedVersions("DROP CONSTRAINT ON ()-[r:TYPE]-() ASSERT EXISTS (r.prop)", DEPRECATED_DROP_CONSTRAINT_SYNTAX)
  }

  test("deprecated create node property existence constraint syntax - deprecate version 0") {
    assertNotificationInSupportedVersions("CREATE CONSTRAINT ON (n:Label) ASSERT EXISTS (n.prop)",
      DEPRECATED_CREATE_PROPERTY_EXISTENCE_CONSTRAINT_SYNTAX)
  }

  test("deprecated create relationship property existence constraint syntax - deprecate version 0") {
    assertNotificationInSupportedVersions("CREATE CONSTRAINT ON ()-[r:TYPE]-() ASSERT EXISTS (r.prop)",
      DEPRECATED_CREATE_PROPERTY_EXISTENCE_CONSTRAINT_SYNTAX)
  }

  test("deprecated create node property existence constraint syntax - deprecate version 1") {
    assertNotificationInSupportedVersions_4_X("CREATE CONSTRAINT ON (n:Label) ASSERT (n.prop) IS NOT NULL",
      DEPRECATED_CREATE_CONSTRAINT_ON_ASSERT_SYNTAX)
  }

  test("deprecated create relationship property existence constraint syntax - deprecate version 1") {
    assertNotificationInSupportedVersions_4_X("CREATE CONSTRAINT ON ()-[r:TYPE]-() ASSERT (r.prop) IS NOT NULL",
      DEPRECATED_CREATE_CONSTRAINT_ON_ASSERT_SYNTAX)
  }

  test("deprecated create node key constraint syntax") {
    assertNotificationInSupportedVersions("CREATE CONSTRAINT ON (n:Label) ASSERT (n.prop) IS NODE KEY",
      DEPRECATED_CREATE_CONSTRAINT_ON_ASSERT_SYNTAX)
  }

  test("deprecated create uniqueness constraint syntax") {
    assertNotificationInSupportedVersions("CREATE CONSTRAINT ON (n:Label) ASSERT (n.prop) IS UNIQUE",
      DEPRECATED_CREATE_CONSTRAINT_ON_ASSERT_SYNTAX)
  }

  test("deprecated create constraint backed by btree index syntax") {
    // OPTIONS was introduced in 4.2
    // FOR ... REQUIRE was introduced in 4.4 and can therefore not be used in this test (since it tests 4.3)
    assertNotificationInSupportedVersions_4_X(
      s"""CREATE CONSTRAINT ON (n:Label)
         |ASSERT (n.prop) IS NODE KEY
         |OPTIONS {IndexProvider: '${GenericNativeIndexProvider.DESCRIPTOR.name()}'}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
    assertNotificationInSupportedVersions_4_X(
      s"""CREATE CONSTRAINT ON (n:Label)
         |ASSERT (n.prop) IS UNIQUE
         |OPTIONS {indexprovider: '${NativeLuceneFusionIndexProviderFactory30.DESCRIPTOR.name()}'}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
    assertNotificationInSupportedVersions_4_X(
      s"""CREATE CONSTRAINT ON (n:Label)
         |ASSERT (n.prop) IS NODE KEY
         |OPTIONS {IndexConfig: {`${SPATIAL_CARTESIAN_MAX.getSettingName}`: [40, 60]}}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
    assertNotificationInSupportedVersions_4_X(
      s"""CREATE CONSTRAINT ON (n:Label)
         |ASSERT (n.prop) IS UNIQUE
         |OPTIONS {indexconfig: {`${SPATIAL_WGS84_MIN.getSettingName}`: [-40, -60]}}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
    assertNoNotificationInSupportedVersions_4_X(
      s"""CREATE CONSTRAINT ON (n:Label)
         |ASSERT (n.prop) IS UNIQUE
         |OPTIONS {indexconfig: {`${FULLTEXT_EVENTUALLY_CONSISTENT.getSettingName}`: false}}""".stripMargin,
      DEPRECATED_BTREE_INDEX_SYNTAX
    )
  }

  test("deprecated show index syntax") {
    val queries = Seq(
      "SHOW INDEXES BRIEF",
      "SHOW INDEXES BRIEF OUTPUT",
      "SHOW INDEXES VERBOSE",
      "SHOW INDEXES VERBOSE OUTPUT",
    )

    // Note: Show indexes was introduced in Neo4j 4.2
    assertNotificationInSupportedVersions_4_X(queries, DEPRECATED_SHOW_SCHEMA_SYNTAX)
  }

  test("deprecated show btree index syntax") {
    // Note: Show indexes was introduced in Neo4j 4.2
    assertNotificationInSupportedVersions_4_X("SHOW BTREE INDEXES", DEPRECATED_BTREE_INDEX_SYNTAX)
  }

  test("deprecated show constraint syntax") {
    val queries = Seq(
      "SHOW CONSTRAINTS BRIEF",
      "SHOW CONSTRAINTS BRIEF OUTPUT",
      "SHOW CONSTRAINTS VERBOSE",
      "SHOW CONSTRAINTS VERBOSE OUTPUT",
    )

    // Note: Show constraints was introduced in Neo4j 4.2
    assertNotificationInSupportedVersions_4_X(queries, DEPRECATED_SHOW_SCHEMA_SYNTAX)
  }

  test("deprecated show existence constraint syntax") {
    val queries = Seq(
      "SHOW EXISTS CONSTRAINT",
      "SHOW NODE EXISTS CONSTRAINT",
      "SHOW RELATIONSHIP EXISTS CONSTRAINT",
    )

    // Note: Show constraints was introduced in Neo4j 4.2
    assertNotificationInSupportedVersions_4_X(queries, DEPRECATED_SHOW_EXISTENCE_CONSTRAINT_SYNTAX)
  }

  // OTHER DEPRECATIONS IN 4.X

  test("deprecated octal literal syntax") {
    assertNotificationInSupportedVersions("RETURN 0123 AS octal", DEPRECATED_OCTAL_LITERAL_SYNTAX)
  }

  test("deprecated hex literal syntax") {
    assertNotificationInSupportedVersions("RETURN 0X12B AS hex", DEPRECATED_HEX_LITERAL_SYNTAX)
  }

  test("deprecated pattern expression syntax") {
    val queries = Seq(
      "MATCH (a) RETURN (a)--()",
      "MATCH (a) WHERE ANY (x IN (a)--() WHERE 1=1) RETURN a",
    )

    assertNotificationInSupportedVersions(queries, DEPRECATED_USE_OF_PATTERN_EXPRESSION)
  }

  test("not deprecated pattern expression syntax") {
    // Existential subqueries was introduced in Neo4j 4.0
    assertNoNotificationInSupportedVersions_4_X("MATCH (a) WHERE EXISTS {(x) WHERE (x)--()} RETURN a", DEPRECATED_USE_OF_PATTERN_EXPRESSION)

    val queries = Seq(
      "MATCH (a)--() RETURN a",
      "MATCH (a) WHERE exists((a)--()) RETURN a",
      "MATCH (a) WHERE (a)--() RETURN a",
      "MATCH (a) RETURN [p=(a)--(b) | p]",
      "RETURN NOT exists(()--())",
      "RETURN NOT ()--()",
      "RETURN ()--() OR ()--()--()",
      """
        |EXPLAIN
        |MATCH (actor:Actor)
        |RETURN actor,
        |  CASE
        |    WHEN (actor)-[:WON]->(:Oscar) THEN 'Oscar winner'
        |    WHEN (actor)-[:WON]->(:GoldenGlobe) THEN 'Golden Globe winner'
        |    ELSE 'None'
        |  END AS accolade
        |""".stripMargin,
      """
        |EXPLAIN
        |MATCH (movie:Movie)<-[:ACTED_IN]-(actor:Actor)
        |WITH movie, collect(actor) AS cast
        |WHERE ANY(actor IN cast WHERE (actor)-[:WON]->(:Award))
        |RETURN movie
        |""".stripMargin
    )
    assertNoNotificationInSupportedVersions(queries, DEPRECATED_USE_OF_PATTERN_EXPRESSION)
  }

  test("deprecated property existence syntax") {
    val queries = Seq(
      // On node
      "MATCH (n) WHERE EXISTS(n.prop) RETURN n",

      // On node with not
      "MATCH (n) WHERE NOT EXISTS(n.prop) RETURN n",

      // On relationship
      "MATCH ()-[r]-() WITH r WHERE EXISTS(r.prop) RETURN r.prop",

      // On map key
      "WITH {key:'blah'} as map RETURN EXISTS(map.key)",

      // On map notation in WHERE
      "MATCH (n) WHERE EXISTS(n['prop']) RETURN n",

      // On map notation in WHERE with NOT
      "MATCH (n) WHERE NOT EXISTS(n['prop']) RETURN n",

      // On map notation in RETURN
      "MATCH (n) RETURN EXISTS(n['prop'])"
    )

    assertNotificationInSupportedVersions(queries, DEPRECATED_PROPERTY_EXISTENCE_SYNTAX)
  }

  test("exists on paths should not be deprecated") {
    assertNoNotificationInSupportedVersions("MATCH (n) WHERE EXISTS( (n)-[:REL]->() ) RETURN count(n)", DEPRECATED_PROPERTY_EXISTENCE_SYNTAX)
  }

  test("exists subclause should not be deprecated") {
    // Note: Exists subclause was introduced in Neo4j 4.0
    assertNoNotificationInSupportedVersions_4_X("MATCH (n) WHERE EXISTS { MATCH (n)-[]->() } RETURN n.prop",
      DEPRECATED_PROPERTY_EXISTENCE_SYNTAX)
  }

  test("deprecated coercion list to boolean") {
    val queries = Seq(
      "RETURN NOT []",
      "RETURN NOT [1]",
      "RETURN NOT ['a']",
      "RETURN ['a'] OR []",
      "RETURN TRUE OR []",
      "RETURN NOT (TRUE OR [])",
      "RETURN ['a'] AND []",
      "RETURN TRUE AND []",
      "RETURN NOT (TRUE AND [])",
      "MATCH (n) WHERE [] RETURN TRUE",
      "MATCH (n) WHERE range(0, 10) RETURN TRUE",
      "MATCH (n) WHERE range(0, 10) RETURN range(0, 10)",
    )

    assertNotificationInSupportedVersions(queries, DEPRECATED_COERCION_OF_LIST_TO_BOOLEAN)

    assertNoNotificationInSupportedVersions("RETURN NOT TRUE", DEPRECATED_COERCION_OF_LIST_TO_BOOLEAN)
  }

  test("should not deprecate boolean coercion of pattern expressions") {
    val queries = Seq(
      "RETURN NOT ()--()",
      "RETURN ()--() OR ()--()--()",
      "MATCH (n) WHERE (n)-[]->() RETURN n",
      """
      |MATCH (a), (b)
      |WITH a, b
      |WHERE a.id = 0
      |  AND (a)-[:T]->(b:Label1)
      |  OR (a)-[:T*]->(b:Label2)
      |RETURN DISTINCT b
      """.stripMargin,
      """
        |MATCH (a), (b)
        |WITH a, b
        |WHERE a.id = 0
        |  AND exists((a)-[:T]->(b:Label1))
        |  OR exists((a)-[:T*]->(b:Label2))
        |RETURN DISTINCT b
      """.stripMargin,
      "MATCH (n) WHERE NOT (n)-[:REL2]-() RETURN n",
      "MATCH (n) WHERE (n)-[:REL1]-() AND (n)-[:REL3]-() RETURN n",
      "MATCH (n WHERE (n)--()) RETURN n",
      """
        |MATCH (actor:Actor)
        |RETURN actor,
        |  CASE
        |    WHEN (actor)-[:WON]->(:Oscar) THEN 'Oscar winner'
        |    WHEN (actor)-[:WON]->(:GoldenGlobe) THEN 'Golden Globe winner'
        |    ELSE 'None'
        |  END AS accolade
        |""".stripMargin,
      """
        |MATCH (movie:Movie)<-[:ACTED_IN]-(actor:Actor)
        |WITH movie, collect(actor) AS cast
        |WHERE ANY(actor IN cast WHERE (actor)-[:WON]->(:Award))
        |RETURN movie
        |""".stripMargin,
    )
    assertNoDeprecations(queries)
  }

  test("should not allow referencing elements being created by the pattern within that same pattern") {
    val badQueries = Seq(
      "CREATE (a {prop:7})-[r:R {prop: a.prop}]->(b)",
      "CREATE (a {prop:a.prop})",
      "CREATE (a {prop:7})-[r:R]->(b {prop: a.prop})",
      "CREATE (a {prop: r.prop})-[r:R {prop:7}]->(b)",
      "CREATE (a)-[r:R {prop:7}]->(b {prop: r.prop})",
      "CREATE (a {prop:b.prop})-[r:R]->(b {prop: 7})",
      "CREATE (a)-[r:R {prop:b.prop}]->(b {prop: 7})",
      "CREATE (a:A:B)-[r:R]->(b {prop: labels(a)})",
      "CREATE p=(a {prop:7})-[r:R {prop: a.prop}]->(b)",
      "CREATE p=(a {prop:7})-[:R]->(b {prop: nodes(p)[0].prop})",
      "CREATE (a {prop:7})-[:R {prop: a.prop}]->(b)",
      "CREATE (a {prop:7})-[r:R]->({prop: a.prop})",
      "CREATE p=({prop:7})-[:R]->(b {prop: nodes(p)[0].prop})",
      "CREATE (n {prop: true IN [a IN [false] WHERE n.prop | a]})",
      "CREATE (n)-[r:R {prop: true IN [a IN [false] WHERE n.prop | a]}]->(m)",
      "CREATE (n)-[r:R {prop: true IN [a IN [false] WHERE r.prop | a]}]->(m)",
      "CREATE (n)-[r:R]->(m {prop: true IN [a IN [false] WHERE n.prop | a]})",
      "CREATE (n)-[r:R]->(m {prop: true IN [a IN [false] WHERE r.prop | a]})",
    )

    assertNotificationInSupportedVersions(badQueries, DEPRECATED_SELF_REFERENCE_TO_VARIABLE_IN_CREATE_PATTERN)

    val okQueries = Seq(
      "CREATE (a)-[:R]->(a)",
      "CREATE (a {prop: 7})-[:R]->(a)",
      "MATCH (b) CREATE ()-[:R]->(a {prop: b.prop})",
      "CREATE (a {prop: 7}) CREATE (b {prop: a.prop})",
      "MATCH (a {prop:7})-[r:R {prop: a.prop}]->(b) RETURN *",
      "CREATE (a {prop:7}) CREATE (a)-[:R]->(b {prop: a.prop})",
      "MATCH (a) CREATE (a)-[:R]->(b {prop: a.prop})",
      "CREATE (n {prop: true IN [n IN [false] | true]})",
      "CREATE (n {prop: true IN [n IN [false] | n]})",
      "CREATE (a)-[r:R {prop: true IN [r IN [false] | true]}]->(b)",
      "CREATE (a)-[r:R {prop: true IN [r IN [false] | r]}]->(b)",
      "CREATE (a)-[r:R {prop: true IN [a IN [false] | a]}]->(b)",
      "CREATE (a)-[r:R]->(b {prop: true IN [r IN [false] | r]})",
      "CREATE (a)-[r:R]->(b {prop: true IN [a IN [false] | a]})"
    )

    assertNoNotificationInSupportedVersions(okQueries, DEPRECATED_SELF_REFERENCE_TO_VARIABLE_IN_CREATE_PATTERN)
  }

  test("deprecated periodic commit hint") {
    val query = "USING PERIODIC COMMIT LOAD CSV FROM 'file:///artists.csv' AS line CREATE (:Artist {name: line[1], year: toInteger(line[2])})"
    assertNotificationInSupportedVersions(query, DEPRECATED_PERIODIC_COMMIT)
  }

  test("use point.distance rather than distance") {
    val deprecated = Seq(" MATCH (n) WHERE distance(n.prop, point({x:0, y:0})) < 10 RETURN n")
    val newSyntax = Seq(" MATCH (n) WHERE point.distance(n.prop, point({x:0, y:0})) < 10 RETURN n")
    val detail = NotificationDetail.Factory.deprecatedName("distance", "point.distance")

    assertNotificationInSupportedVersions(deprecated, DEPRECATED_FUNCTION, detail)
    assertNoNotificationInSupportedVersions(newSyntax, DEPRECATED_FUNCTION)
  }

  test("comparing points is deprecated") {
    val deprecated = Seq(
      "MATCH (n) WHERE n.prop < point({x:0, y:0}) RETURN n",
      "MATCH (n) WHERE n.prop <= point({x:0, y:0}) RETURN n",
      "MATCH (n) WHERE n.prop > point({x:0, y:0}) RETURN n",
      "MATCH (n) WHERE n.prop >= point({x:0, y:0}) RETURN n",
      "RETURN point({x:0, y:0}) < point({x:0, y:0}) <= point({x:0, y:0})"
    )
    assertNotificationInSupportedVersions(deprecated, DEPRECATED_POINTS_COMPARE)
  }

  test("btree index hint") {
    val deprecated = Seq(
      "MATCH (n:Label) USING BTREE INDEX n:Label(prop) WHERE n.prop > 5 RETURN n",
      "MATCH (n:Label) USING BTREE INDEX SEEK n:Label(prop) WHERE n.prop > 5 RETURN n",
      "MATCH ()-[r:REL]->() USING BTREE INDEX r:REL(prop) WHERE r.prop > 5 RETURN r",
      "MATCH ()-[r:REL]->() USING BTREE INDEX SEEK r:REL(prop) WHERE r.prop > 5 RETURN r",
    )

    val notDeprecated = Seq(
      "MATCH (n:Label) USING INDEX n:Label(prop) WHERE n.prop > 5 RETURN n",
      "MATCH ()-[r:REL]->() USING TEXT INDEX r:REL(prop) WHERE r.prop > 5 RETURN r",
      "MATCH (n:Label) USING TEXT INDEX SEEK n:Label(prop) WHERE n.prop > 5 RETURN n",
      "MATCH ()-[r:REL]->() USING INDEX SEEK r:REL(prop) WHERE r.prop > 5 RETURN r",
    )

    assertNotificationInSupportedVersions(deprecated, DEPRECATED_BTREE_INDEX_SYNTAX)

    assertNoNotificationInSupportedVersions(notDeprecated, DEPRECATED_BTREE_INDEX_SYNTAX)
  }

  test("deprecated case when null syntax") {
    assertNotificationInSupportedVersions("MATCH (n) RETURN CASE n WHEN null THEN 1 ELSE 2 END", DEPRECATED_CASE_EXPRESSION)
  }

  // FUNCTIONALITY DEPRECATED IN 3.5, REMOVED IN 4.0

  test("deprecated toInt") {
    val query = "RETURN toInt('1') AS one"
    val detail = NotificationDetail.Factory.deprecatedName("toInt", "toInteger")
    assertNotificationInLastMajorVersion(query, DEPRECATED_FUNCTION, detail)
  }

  test("deprecated upper") {
    val query = "RETURN upper('foo') AS upper"
    val detail = NotificationDetail.Factory.deprecatedName("upper", "toUpper")
    assertNotificationInLastMajorVersion(query, DEPRECATED_FUNCTION, detail)
  }

  test("deprecated lower") {
    val query = "RETURN lower('BAR') AS lower"
    val detail = NotificationDetail.Factory.deprecatedName("lower", "toLower")
    assertNotificationInLastMajorVersion(query, DEPRECATED_FUNCTION, detail)
  }

  test("deprecated rels") {
    val query = "MATCH p = ()-->() RETURN rels(p) AS r"
    val detail = NotificationDetail.Factory.deprecatedName("rels", "relationships")
    assertNotificationInLastMajorVersion(query, DEPRECATED_FUNCTION, detail)
  }

  test("deprecated filter") {
    val query = "WITH [1,2,3] AS list RETURN filter(x IN list WHERE x % 2 = 1) AS odds"
    val detail = NotificationDetail.Factory.deprecatedName("filter(...)", "[...]")
    assertNotificationInLastMajorVersion(query, DEPRECATED_FUNCTION, detail)
  }

  test("deprecated extract") {
    val query = "WITH [1,2,3] AS list RETURN extract(x IN list | x * 10) AS tens"
    val detail = NotificationDetail.Factory.deprecatedName("extract(...)", "[...]")
    assertNotificationInLastMajorVersion(query, DEPRECATED_FUNCTION, detail)
  }

  test("deprecated parameter syntax") {
    assertNotificationInLastMajorVersion("RETURN {param} AS parameter", DEPRECATED_PARAMETER_SYNTAX)
  }

  test("deprecated parameter syntax for property map") {
    assertNotificationInLastMajorVersion("CREATE (:Label {props})", DEPRECATED_PARAMETER_SYNTAX)
  }

  test("deprecated length of string") {
    assertNotificationInLastMajorVersion("RETURN length('a string')", LENGTH_ON_NON_PATH)
  }

  test("deprecated length of list") {
    assertNotificationInLastMajorVersion("RETURN length([1, 2, 3])", LENGTH_ON_NON_PATH)
  }

  test("deprecated length of pattern expression") {
    assertNotificationInLastMajorVersion("MATCH (a) WHERE a.name='Alice' RETURN length((a)-->()-->())", LENGTH_ON_NON_PATH)
  }

  test("deprecated future ambiguous reltype separator") {
    val queries = Seq(
      "MATCH (a)-[:A|:B|:C {foo:'bar'}]-(b) RETURN a,b",
      "MATCH (a)-[x:A|:B|:C]-() RETURN a",
      "MATCH (a)-[:A|:B|:C*]-() RETURN a"
    )

    assertNotificationInLastMajorVersion(queries, DEPRECATED_RELATIONSHIP_TYPE_SEPARATOR)

    // clear caches of the rewritten queries to not keep notifications around
    dbms.clearQueryCaches()
  }

  test("not deprecated reltype separator cases") {
    val queries = Seq(
      "MATCH (a)-[:A|B|C {foo:'bar'}]-(b) RETURN a,b",
      "MATCH (a)-[:A|:B|:C]-(b) RETURN a,b",
      "MATCH (a)-[:A|B|C]-(b) RETURN a,b"
    )

    assertNoNotificationInLastMajorVersion(queries, DEPRECATED_RELATIONSHIP_TYPE_SEPARATOR)
  }
}
